@@ -32,53 +32,11 @@ typedef struct {
     ucp_proto_rndv_rtr_data_received_cb_t data_received;
 } ucp_proto_rndv_rtr_priv_t;
 
-
-static ucs_status_t
-ucp_proto_rndv_rtr_common_init(const ucp_proto_init_params_t *init_params,
-                               uint64_t rndv_modes, size_t max_length,
-                               ucs_linear_func_t unpack_time,
-                               ucp_proto_perf_node_t *unpack_perf_node,
-                               ucp_md_map_t md_map,
-                               ucs_memory_type_t mem_type,
-                               ucs_sys_device_t sys_dev)
-{
-    ucp_context_h context                    = init_params->worker->context;
-    ucp_proto_rndv_ctrl_init_params_t params = {
-        .super.super         = *init_params,
-        .super.latency       = 0,
-        .super.overhead      = 40e-9,
-        .super.cfg_thresh    = ucp_proto_rndv_cfg_thresh(context, rndv_modes),
-        .super.cfg_priority  = 0,
-        .super.min_length    = 1,
-        .super.max_length    = max_length,
-        .super.min_frag_offs = UCP_PROTO_COMMON_OFFSET_INVALID,
-        .super.max_frag_offs = ucs_offsetof(uct_iface_attr_t, cap.am.max_bcopy),
-        .super.max_iov_offs  = UCP_PROTO_COMMON_OFFSET_INVALID,
-        .super.hdr_size      = sizeof(ucp_rndv_rtr_hdr_t),
-        .super.send_op       = UCT_EP_OP_AM_BCOPY,
-        .super.memtype_op    = UCT_EP_OP_LAST,
-        .super.flags         = UCP_PROTO_COMMON_INIT_FLAG_RESPONSE |
-                               UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
-        .super.exclude_map   = 0,
-        .remote_op_id        = UCP_OP_ID_RNDV_SEND,
-        .unpack_time         = unpack_time,
-        .unpack_perf_node    = unpack_perf_node,
-        .perf_bias           = 0.0,
-        .mem_info.type       = mem_type,
-        .mem_info.sys_dev    = sys_dev,
-        .ctrl_msg_name       = UCP_PROTO_RNDV_RTR_NAME,
-        .md_map              = md_map
-    };
-    ucs_status_t status;
-
-    status = ucp_proto_rndv_ctrl_am_init(&params);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    *init_params->priv_size = sizeof(ucp_proto_rndv_rtr_priv_t);
-    return UCS_OK;
-}
+typedef struct {
+    ucp_proto_rndv_rtr_priv_t super;
+    ucs_memory_type_t         frag_mem_type;
+    ucs_sys_device_t          frag_sys_dev;
+} ucp_proto_rndv_rtr_mtype_priv_t;
 
 static UCS_F_ALWAYS_INLINE void
 ucp_proto_rtr_common_request_init(ucp_request_t *req)
@@ -99,10 +57,6 @@ static ucs_status_t ucp_proto_rndv_rtr_common_send(ucp_request_t *req)
                                                       rpriv->super.lane,
                                                       rpriv->pack_cb, req,
                                                       max_rtr_size, NULL, 0);
-    if (status == UCS_OK) {
-        UCP_WORKER_STAT_RNDV(worker, RTR, +1);
-    }
-
     return status;
 }
 
@@ -173,12 +127,13 @@ static size_t ucp_proto_rndv_rtr_pack_with_rkey(void *dest, void *arg)
 
 static ucs_status_t ucp_proto_rndv_rtr_progress(uct_pending_req_t *self)
 {
-    ucp_request_t *req = ucs_container_of(self, ucp_request_t, send.uct);
+    ucp_request_t *req   = ucs_container_of(self, ucp_request_t, send.uct);
+    ucp_worker_t *worker = req->send.ep->worker;
     const ucp_proto_rndv_rtr_priv_t *rpriv = req->send.proto_config->priv;
     ucs_status_t status;
 
     if (!(req->flags & UCP_REQUEST_FLAG_PROTO_INITIALIZED)) {
-        status = ucp_datatype_iter_mem_reg(req->send.ep->worker->context,
+        status = ucp_datatype_iter_mem_reg(worker->context,
                                            &req->send.state.dt_iter,
                                            rpriv->super.md_map,
                                            UCT_MD_MEM_ACCESS_REMOTE_PUT |
@@ -193,38 +148,57 @@ static ucs_status_t ucp_proto_rndv_rtr_progress(uct_pending_req_t *self)
         req->flags |= UCP_REQUEST_FLAG_PROTO_INITIALIZED;
     }
 
-    return ucp_proto_rndv_rtr_common_send(req);
+    status = ucp_proto_rndv_rtr_common_send(req);
+    if (status == UCS_OK) {
+        UCP_WORKER_STAT_RNDV(worker, RTR, +1);
+    }
+
+    return status;
 }
 
-static ucs_status_t
-ucp_proto_rndv_rtr_init(const ucp_proto_init_params_t *init_params)
+static void ucp_proto_rndv_rtr_probe(const ucp_proto_init_params_t *init_params)
 {
-    const uint64_t rndv_modes        = UCS_BIT(UCP_RNDV_MODE_PUT_ZCOPY) |
-                                       UCS_BIT(UCP_RNDV_MODE_AM);
-    ucp_proto_rndv_rtr_priv_t *rpriv = init_params->priv;
-    ucs_status_t status;
+    ucp_context_h context                   = init_params->worker->context;
+    ucp_proto_rndv_ctrl_init_params_t params = {
+        .super.super         = *init_params,
+        .super.latency       = 0,
+        .super.overhead      = context->config.ext.proto_overhead_rndv_rtr,
+        .super.cfg_thresh    = ucp_proto_rndv_cfg_thresh(context,
+                               UCS_BIT(UCP_RNDV_MODE_PUT_ZCOPY) |
+                               UCS_BIT(UCP_RNDV_MODE_AM)),
+        .super.cfg_priority  = 80,
+        .super.min_length    = 1,
+        .super.max_length    = SIZE_MAX,
+        .super.min_frag_offs = UCP_PROTO_COMMON_OFFSET_INVALID,
+        .super.max_frag_offs = ucs_offsetof(uct_iface_attr_t, cap.am.max_bcopy),
+        .super.max_iov_offs  = UCP_PROTO_COMMON_OFFSET_INVALID,
+        .super.hdr_size      = sizeof(ucp_rndv_rtr_hdr_t),
+        .super.send_op       = UCT_EP_OP_AM_BCOPY,
+        .super.memtype_op    = UCT_EP_OP_LAST,
+        .super.flags         = UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING,
+        .super.exclude_map   = 0,
+        .super.reg_mem_info  = ucp_proto_common_select_param_mem_info(
+                                                     init_params->select_param),
+        .remote_op_id        = UCP_OP_ID_RNDV_SEND,
+        .lane                = ucp_proto_rndv_find_ctrl_lane(init_params),
+        .perf_bias           = 0.0,
+        .ctrl_msg_name       = UCP_PROTO_RNDV_RTR_NAME,
+        .md_map              = 0
+    };
+    ucp_proto_rndv_rtr_priv_t rpriv;
 
     if (!ucp_proto_rndv_op_check(init_params, UCP_OP_ID_RNDV_RECV, 0)) {
-        return UCS_ERR_UNSUPPORTED;
+        return;
     }
 
-    status = ucp_proto_rndv_rtr_common_init(init_params, rndv_modes, SIZE_MAX,
-                                            UCS_LINEAR_FUNC_ZERO, NULL, 0,
-                                            init_params->select_param->mem_type,
-                                            init_params->select_param->sys_dev);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    rpriv->data_received = ucp_proto_rndv_rtr_data_received;
-
     if (init_params->select_param->dt_class == UCP_DATATYPE_CONTIG) {
-        rpriv->pack_cb = ucp_proto_rndv_rtr_pack_with_rkey;
+        rpriv.pack_cb = ucp_proto_rndv_rtr_pack_with_rkey;
     } else {
-        rpriv->pack_cb = ucp_proto_rndv_rtr_pack_without_rkey;
+        rpriv.pack_cb = ucp_proto_rndv_rtr_pack_without_rkey;
     }
+    rpriv.data_received = ucp_proto_rndv_rtr_data_received;
 
-    return UCS_OK;
+    ucp_proto_rndv_ctrl_probe(&params, &rpriv, sizeof(rpriv));
 }
 
 static void ucp_proto_rndv_rtr_query(const ucp_proto_query_params_t *params,
@@ -232,8 +206,8 @@ static void ucp_proto_rndv_rtr_query(const ucp_proto_query_params_t *params,
 {
     const ucp_proto_rndv_ctrl_priv_t *rpriv = params->priv;
 
-    ucp_proto_select_elem_query(params->worker, &rpriv->remote_proto,
-                                params->msg_length, attr);
+    ucp_proto_config_query(params->worker, &rpriv->remote_proto_config,
+                           params->msg_length, attr);
     attr->is_estimation = 1;
     attr->lane_map      = UCS_BIT(rpriv->lane);
 }
@@ -268,7 +242,7 @@ ucp_proto_t ucp_rndv_rtr_proto = {
     .name     = "rndv/rtr",
     .desc     = NULL,
     .flags    = 0,
-    .init     = ucp_proto_rndv_rtr_init,
+    .probe    = ucp_proto_rndv_rtr_probe,
     .query    = ucp_proto_rndv_rtr_query,
     .progress = {ucp_proto_rndv_rtr_progress},
     .abort    = ucp_proto_rndv_rtr_abort,
@@ -306,11 +280,14 @@ static size_t ucp_proto_rndv_rtr_mtype_pack(void *dest, void *arg)
     return sizeof(*rtr) + packed_rkey_size;
 }
 
-static void ucp_proto_rndv_rtr_mtype_complete(ucp_request_t *req)
+static UCS_F_ALWAYS_INLINE void
+ucp_proto_rndv_rtr_mtype_complete(ucp_request_t *req, int abort)
 {
-    ucs_mpool_put_inline(req->send.rndv.mdesc);
+    if (!abort || (req->send.rndv.mdesc != NULL)) {
+        ucs_mpool_put_inline(req->send.rndv.mdesc);
+    }
     if (ucp_proto_rndv_request_is_ppln_frag(req)) {
-        ucp_proto_rndv_ppln_recv_frag_complete(req, 0, 0);
+        ucp_proto_rndv_ppln_recv_frag_complete(req, 0, abort);
     } else {
         ucp_proto_rndv_rtr_common_complete(req, UCS_BIT(UCP_DATATYPE_CONTIG));
     }
@@ -319,24 +296,26 @@ static void ucp_proto_rndv_rtr_mtype_complete(ucp_request_t *req)
 static void
 ucp_proto_rndv_rtr_mtype_abort(ucp_request_t *req, ucs_status_t status)
 {
-    req->status = status;
-    ucp_send_request_id_release(req);
-    ucs_mpool_put_inline(req->send.rndv.mdesc);
+    ucp_request_t *super_req = ucp_request_get_super(req);
 
+    super_req->status = status;
+
+    /* When pipeline is used, there is a top-level 'recv' request that we also
+     * need to abort */
     if (ucp_proto_rndv_request_is_ppln_frag(req)) {
-        ucp_proto_rndv_ppln_recv_frag_complete(req, 0, 1);
-    } else {
-        ucp_proto_rndv_rtr_common_complete(req, UCS_BIT(UCP_DATATYPE_CONTIG));
+        ucp_request_get_super(super_req)->status = status;
     }
+
+    /*TODO: Invalidate memh */
+    ucp_send_request_id_release(req);
+    ucp_proto_rndv_rtr_mtype_complete(req, 1);
 }
 
 static ucs_status_t ucp_proto_rndv_rtr_mtype_reset(ucp_request_t *req)
 {
-    ucs_mpool_put_inline(req->send.rndv.mdesc);
-    if (ucp_proto_rndv_request_is_ppln_frag(req)) {
-        req->status = UCS_ERR_CANCELED;
-        ucp_proto_rndv_ppln_recv_frag_clean(req);
-        return UCS_ERR_CANCELED;
+    if (req->flags & UCP_REQUEST_FLAG_PROTO_INITIALIZED) {
+        ucs_mpool_put_inline(req->send.rndv.mdesc);
+        req->send.rndv.mdesc = NULL;
     }
 
     return ucp_proto_request_zcopy_id_reset(req);
@@ -346,7 +325,7 @@ static void ucp_proto_rndv_rtr_mtype_copy_completion(uct_completion_t *uct_comp)
 {
     ucp_request_t *req = ucs_container_of(uct_comp, ucp_request_t,
                                           send.state.uct_comp);
-    ucp_proto_rndv_rtr_mtype_complete(req);
+    ucp_proto_rndv_rtr_mtype_complete(req, 0);
 }
 
 static void
@@ -356,25 +335,25 @@ ucp_proto_rndv_rtr_mtype_data_received(ucp_request_t *req, int in_buffer)
     if (in_buffer) {
         /* Data was already placed in user buffer because the sender responded
            with RNDV_DATA packets */
-        ucp_proto_rndv_rtr_mtype_complete(req);
+        ucp_proto_rndv_rtr_mtype_complete(req, 0);
     } else {
         /* Data was not placed in user buffer, which means it was placed to
            the remote address we published - the rendezvous fragment */
-        ucp_proto_rndv_mtype_copy(req, req->send.rndv.mdesc->ptr,
-                                  ucp_proto_rndv_mtype_get_req_memh(req),
-                                  uct_ep_put_zcopy,
-                                  ucp_proto_rndv_rtr_mtype_copy_completion,
-                                  "out to");
+        ucp_proto_rndv_mdesc_mtype_copy(
+                req, uct_ep_put_zcopy, ucp_proto_rndv_rtr_mtype_copy_completion,
+                "out to");
     }
 }
 
 static ucs_status_t ucp_proto_rndv_rtr_mtype_progress(uct_pending_req_t *self)
 {
     ucp_request_t *req = ucs_container_of(self, ucp_request_t, send.uct);
+    const ucp_proto_rndv_rtr_mtype_priv_t *rpriv = req->send.proto_config->priv;
     ucs_status_t status;
 
     if (!(req->flags & UCP_REQUEST_FLAG_PROTO_INITIALIZED)) {
-        status = ucp_proto_rndv_mtype_request_init(req);
+        status = ucp_proto_rndv_mtype_request_init(req, rpriv->frag_mem_type,
+                                                   rpriv->frag_sys_dev);
         if (status != UCS_OK) {
             ucp_proto_request_abort(req, status);
             return UCS_OK;
@@ -384,74 +363,114 @@ static ucs_status_t ucp_proto_rndv_rtr_mtype_progress(uct_pending_req_t *self)
         req->flags |= UCP_REQUEST_FLAG_PROTO_INITIALIZED;
     }
 
-    return ucp_proto_rndv_rtr_common_send(req);
-}
-
-static ucs_status_t
-ucp_proto_rndv_rtr_mtype_init(const ucp_proto_init_params_t *init_params)
-{
-    ucp_proto_rndv_rtr_priv_t *rpriv = init_params->priv;
-    const uint64_t rndv_modes        = UCS_BIT(UCP_RNDV_MODE_PUT_PIPELINE);
-    ucp_context_h context            = init_params->worker->context;
-    ucp_proto_perf_node_t *unpack_perf_node;
-    ucs_linear_func_t unpack_time;
-    ucp_md_map_t md_map, dummy_md_map;
-    ucs_status_t status;
-    size_t frag_size;
-    ucp_md_index_t md_index;
-
-    if (!ucp_proto_rndv_op_check(init_params, UCP_OP_ID_RNDV_RECV, 1)) {
-        return UCS_ERR_UNSUPPORTED;
+    status = ucp_proto_rndv_rtr_common_send(req);
+    if (status == UCS_OK) {
+        UCP_WORKER_STAT_RNDV(req->send.ep->worker, RTR_MTYPE, +1);
     }
 
-    status = ucp_proto_rndv_mtype_init(init_params, &dummy_md_map, &frag_size);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    status = ucp_proto_init_buffer_copy_time(
-            init_params->worker, "rtr/mtype unpack", UCS_MEMORY_TYPE_HOST,
-            init_params->select_param->mem_type, UCT_EP_OP_PUT_ZCOPY,
-            &unpack_time, &unpack_perf_node);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    /* Make sure that key of the md used to allocate a bounce buffer will be
-     * packed to the RTR rkey
-     */
-    status = ucp_mm_get_alloc_md_index(context, &md_index);
-    if (status != UCS_OK) {
-        return status;
-    }
-
-    md_map = UCS_BIT(md_index);
-    status = ucp_proto_rndv_rtr_common_init(init_params, rndv_modes, frag_size,
-                                            unpack_time, unpack_perf_node,
-                                            md_map, UCS_MEMORY_TYPE_HOST,
-                                            UCS_SYS_DEVICE_ID_UNKNOWN);
-    ucp_proto_perf_node_deref(&unpack_perf_node);
-
-    rpriv->pack_cb       = ucp_proto_rndv_rtr_mtype_pack;
-    rpriv->data_received = ucp_proto_rndv_rtr_mtype_data_received;
-
     return status;
 }
 
+static void
+ucp_proto_rndv_rtr_mtype_probe(const ucp_proto_init_params_t *init_params)
+{
+    ucp_context_h context                    = init_params->worker->context;
+    ucp_proto_rndv_ctrl_init_params_t params = {
+        .super.super         = *init_params,
+        .super.latency       = 0,
+        .super.overhead      = context->config.ext.proto_overhead_rndv_rtr,
+        .super.cfg_thresh    = ucp_proto_rndv_cfg_thresh(context,
+                               UCS_BIT(UCP_RNDV_MODE_PUT_PIPELINE)),
+        .super.cfg_priority  = 80,
+        .super.min_length    = 1,
+        .super.min_frag_offs = UCP_PROTO_COMMON_OFFSET_INVALID,
+        .super.max_frag_offs = ucs_offsetof(uct_iface_attr_t, cap.am.max_bcopy),
+        .super.max_iov_offs  = UCP_PROTO_COMMON_OFFSET_INVALID,
+        .super.hdr_size      = sizeof(ucp_rndv_rtr_hdr_t),
+        .super.send_op       = UCT_EP_OP_AM_BCOPY,
+        .super.memtype_op    = UCT_EP_OP_LAST,
+        .super.flags         = UCP_PROTO_COMMON_INIT_FLAG_ERR_HANDLING |
+                               UCP_PROTO_COMMON_KEEP_MD_MAP,
+        .super.exclude_map   = 0,
+        .remote_op_id        = UCP_OP_ID_RNDV_SEND,
+        .lane                = ucp_proto_rndv_find_ctrl_lane(init_params),
+        .perf_bias           = 0.0,
+        .ctrl_msg_name       = UCP_PROTO_RNDV_RTR_NAME,
+    };
+    ucs_memory_type_t frag_mem_type;
+    ucp_proto_rndv_rtr_mtype_priv_t rpriv;
+    ucp_md_map_t dummy_md_map;
+    ucp_md_index_t md_index;
+    ucs_status_t status;
+
+    if (!ucp_proto_rndv_op_check(init_params, UCP_OP_ID_RNDV_RECV, 1) ||
+        (init_params->rkey_cfg_index == UCP_WORKER_CFG_INDEX_NULL)) {
+        return;
+    }
+
+    ucs_for_each_bit(frag_mem_type, context->config.ext.rndv_frag_mem_types) {
+        status = ucp_proto_rndv_mtype_init(init_params, frag_mem_type,
+                                           &dummy_md_map,
+                                           &params.super.max_length);
+        if (status != UCS_OK) {
+            continue;
+        }
+
+        status = ucp_mm_get_alloc_md_index(context, frag_mem_type,
+                                           init_params->select_param->sys_dev,
+                                           &md_index,
+                                           &params.super.reg_mem_info);
+        if ((status == UCS_OK) && (md_index != UCP_NULL_RESOURCE)) {
+            params.md_map = UCS_BIT(md_index);
+        } else if (frag_mem_type == UCS_MEMORY_TYPE_HOST) {
+            params.md_map = 0;
+        } else {
+            /* To use non-host staging buffers it should be possible to
+             * allocate them with MD */
+            continue;
+        }
+
+        status = ucp_proto_perf_create("rtr/mtype unpack", &params.unpack_perf);
+        if (status != UCS_OK) {
+            return;
+        }
+
+        status = ucp_proto_init_add_buffer_copy_time(
+                init_params->worker, "unpack copy", frag_mem_type,
+                init_params->select_param->mem_type, UCT_EP_OP_PUT_ZCOPY,
+                params.super.min_length, params.super.max_length, 1,
+                params.unpack_perf);
+        if (status != UCS_OK) {
+            goto out_unpack_perf_destroy;
+        }
+
+        rpriv.super.pack_cb       = ucp_proto_rndv_rtr_mtype_pack;
+        rpriv.super.data_received = ucp_proto_rndv_rtr_mtype_data_received;
+        rpriv.frag_mem_type       = frag_mem_type;
+        rpriv.frag_sys_dev        = params.super.reg_mem_info.sys_dev;
+
+        ucp_proto_rndv_ctrl_probe(&params, &rpriv, sizeof(rpriv));
+out_unpack_perf_destroy:
+        ucp_proto_perf_destroy(params.unpack_perf);
+    }
+}
+
 static void
 ucp_proto_rndv_rtr_mtype_query(const ucp_proto_query_params_t *params,
                                ucp_proto_query_attr_t *attr)
 {
-    const ucp_proto_rndv_ctrl_priv_t *rpriv = params->priv;
+    const ucp_proto_rndv_rtr_mtype_priv_t *rpriv = params->priv;
     ucp_proto_query_attr_t remote_attr;
 
-    ucp_proto_select_elem_query(params->worker, &rpriv->remote_proto,
-                                params->msg_length, &remote_attr);
+    ucp_proto_config_query(params->worker,
+                           &rpriv->super.super.remote_proto_config,
+                           params->msg_length, &remote_attr);
 
     attr->is_estimation  = 1;
     attr->max_msg_length = remote_attr.max_msg_length;
-    attr->lane_map       = UCS_BIT(rpriv->lane);
-    ucp_proto_rndv_mtype_query_desc(params, attr, remote_attr.desc);
+    attr->lane_map       = UCS_BIT(rpriv->super.super.lane);
+    ucp_proto_rndv_mtype_query_desc(params, rpriv->frag_mem_type, attr,
+                                    remote_attr.desc);
     ucs_strncpy_safe(attr->config, remote_attr.config, sizeof(attr->config));
 }
 
@@ -459,7 +478,7 @@ ucp_proto_t ucp_rndv_rtr_mtype_proto = {
     .name     = "rndv/rtr/mtype",
     .desc     = NULL,
     .flags    = 0,
-    .init     = ucp_proto_rndv_rtr_mtype_init,
+    .probe    = ucp_proto_rndv_rtr_mtype_probe,
     .query    = ucp_proto_rndv_rtr_mtype_query,
     .progress = {ucp_proto_rndv_rtr_mtype_progress},
     .abort    = ucp_proto_rndv_rtr_mtype_abort,
